#include <assert.h>
#include <ctype.h>
#include <dirent.h>
#include <fcntl.h>
#include <pwd.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <sys/file.h>
#include <sys/wait.h>
#include "common.h"

static int	days[12] = { 31, 28, 31, 30, 31, 30,
			     31, 31, 30, 31, 30, 31 };

const char	*configFilePath = LIB"/dial-up_utils.config";
const char	*userFilePath = LIB"/dial-up_users";
int		userFileFD;
FILE		*userFile;
const char	*banFilePath = LIB"/banned_users";
int		banFileFD;
FILE		*banFile;

void errQuit(char *format, ...) {
  char buf[256];
  va_list ap;
  va_start(ap, format);
  vsprintf(buf, format, ap);
  printf("%s\n", buf);
  if (userFile) userFileClose();
  if (banFile) banFileClose();
  exit(1);
}

void perrQuit(char *name) {
  perror(name);
  if (userFile) userFileClose();
  if (banFile) banFileClose();
  exit(1);
}

int runCommand(char *path, ...) {
  char *arg[16];
  va_list ap;
  int status, i;
  pid_t pid;
  va_start(ap, path);
  for (i = strlen(path) - 1; i >= 0 && path[i] != '/'; i--);
  arg[0] = path + i + 1;
  for (i = 1; i < 16; i++) {
    arg[i] = va_arg(ap, char*);
    if (arg[i] == NULL) break;
  }
  if ((pid = fork()) < 0)
    perrQuit("fork");
  else if (!pid) {
    execv(path, arg);
    perrQuit("execv");
  } else waitpid(pid, &status, 0);
  return WEXITSTATUS(status);
}

time_t addTime(char *s) {
  time_t res, t = time(NULL), endOfDay, endOfMonth, endOfYear;
  struct tm *tmP = localtime(&t);
  int len = strlen(s);
  char lastC = tolower(s[len - 1]);
  int i, j, n;
  if (!isdigit(lastC)) s[len - 1] = '\0';
  n = atoi(s);
  if (n <= 0) return 0;
  endOfDay = t +
    (60 - tmP->tm_sec) +
    60 * (59 - tmP->tm_min) +
    60 * 60 * (23 - tmP->tm_hour);
  if (isdigit(lastC) || lastC == 'd')
    return endOfDay + (n - 1) * 24 * 60 * 60;
  endOfMonth = endOfDay +
    24 * 60 * 60 * (days[tmP->tm_mon] - 1 - (tmP->tm_mday - 1));
  if (lastC == 'm') {
    for (res = endOfMonth, j = (tmP->tm_mon + 1) % 12, i = 1; i < n; i++) {
      res += days[j] * 24 * 60 * 60;
      j = (j + 1) % 12;
    }
    return res;
  }
  for (endOfYear = endOfMonth, i = tmP->tm_mon + 1; i < 12; i++)
    endOfYear += 24 * 60 * 60 * days[i];
  if (lastC == 'y')
    return endOfYear + 365 * 24 * 60 * 60 * (n - 1);
  return 0;
}

void procList(int *nProcs, ProcRec *procRec) {
  int i, li;
  FILE *f;
  char tpath[32], line[256], token[256];
  DIR *dir = opendir("/proc");
  struct dirent *_dirent;
  if (!dir) perrQuit("opendir");
  *nProcs = 0;
  while ((_dirent = readdir(dir)) != NULL) {
    if (!isdigit(_dirent->d_name[0])) continue;
    procRec[*nProcs].pid = atoi(_dirent->d_name);
    sprintf(tpath, "/proc/%s/stat", _dirent->d_name);
    f = fopen(tpath, "r");
    if (!f) continue;
    if (!fgets(line, 256, f)) goto l0;
    for (li = 0, i = 0; i < 2; i++)
      if (getToken(token, line, &li)) goto l0;
    token[strlen(token) - 1] = '\0';
    strncpy(procRec[*nProcs].name, token + 1, 15);
    procRec[*nProcs].name[15] = '\0';
    for (i = 2; i < 7; i++)
      if (getToken(token, line, &li)) goto l0;
    procRec[*nProcs].tty = atoi(token);
    fclose(f);
    sprintf(tpath, "/proc/%s/status", _dirent->d_name);
    f = fopen(tpath, "r");
    if (!f) continue;
    for (i = 0; i < 5; i++)
      if (!fgets(line, 256, f)) goto l0;
    for (li = 0, i = 0; i < 2; i++)
      if (getToken(token, line, &li)) goto l0;
    procRec[*nProcs].uid = atoi(token);
    (*nProcs)++;
l0:
    fclose(f);
  }
  closedir(dir);
}

int allLinesBusy(int nProcs, ProcRec *procRec, int devBase, int nLines) {
  int i, n;
  byte lineBusy[256];
  bzero(lineBusy, 256);
  for (i = 0; i < nProcs; i++) {
    if (!procRec[i].uid) continue;
    n = procRec[i].tty - devBase;
    if (n >= 0 && n < nLines)
      lineBusy[n] = 1;
  }
  for (i = 0; i < nLines; i++)
    if (!lineBusy[i]) return 0;
  return 1;
}

void userList(int *nLogins, LoginRec *loginRec) {
  struct utmp _utmp;
  uid_t uid;
  char login[9];
  int i;
  FILE *f = fopen("/var/run/utmp", "rb");
  if (!f) f = fopen("/var/adm/utmp", "rb");
  if (!f) perrQuit("fopen");
  *nLogins = 0;
  while (fread(&_utmp, sizeof(struct utmp), 1, f)) {
    if (_utmp.ut_type != USER_PROCESS) continue;
    strncpy(login, _utmp.ut_user, 8);
    for (i = 0; i < 8; i++)
      if (login[i] == 0x20) login[i] = 0;
    login[i] = 0;
    if ((uid = uidFromLogin(login)) != (uid_t)-1)
      loginRec[*nLogins].uid = uid;
    strcpy(loginRec[*nLogins].tty, _utmp.ut_line);
    loginRec[(*nLogins)++].time = _utmp.ut_time;
  }
  fclose(f);
}

int userLoggedIn(int nLogins, LoginRec *loginRec, uid_t uid) {
  int i;
  for (i = 0; i < nLogins; i++) if (loginRec[i].uid == uid) return 1;
  return 0;
}

int getToken(char *d, char *s, int *idx) {
  int si, di, slen = strlen(s);
  for (si = *idx; si < slen && isspace(s[si]); si++);
  for (di = 0; si < slen && isgraph(s[si]); si++) d[di++] = s[si];
  d[di] = '\0';
  *idx = si;
  return di == 0;
}

void fCopy(char *spath, char *dpath) {
  int fd, n;
  char buf[4096];
  FILE *s, *d;
  s = fopen(spath, "rb");
  if (!strcmp(dpath, userFilePath)) {
    fd = open(userFilePath, O_CREAT | O_RDWR | O_TRUNC, 0644);
    assert(fd >= 0);
    d = fdopen(fd, "wb");
    flock(fd, LOCK_EX);
  } else d = fopen(dpath, "wb");
  assert(s != NULL); assert(d != NULL);
  setvbuf(s, NULL, _IOFBF, 4096); setvbuf(d, NULL, _IOFBF, 4096);
  while ((n = fread(buf, 1, 4096, s)) > 0)
    fwrite(buf, 1, n, d);
  fclose(s); fclose(d);
}

uid_t uidFromLogin(char *login) {
  uid_t res = -1;
  struct passwd *passwdp;
  if ((passwdp = getpwnam(login)) != NULL)
    res = passwdp->pw_uid;
  return res;
}

char *loginFromUID(char *login, uid_t uid) {
  char *res = NULL;
  struct passwd *passwdp;
  if ((passwdp = getpwuid(uid)) != NULL) {
    strcpy(login, passwdp->pw_name);
    res = login;
  }
  return res;
}

void userFileOpen() {
  if (!geteuid()) {
    userFileFD = open(userFilePath, O_CREAT | O_RDWR, 0644);
    assert(userFileFD >= 0);
    userFile = fdopen(userFileFD, "r+b");
    flock(userFileFD, LOCK_EX);
  } else {
    userFileFD = open(userFilePath, O_RDONLY);
    assert(userFileFD >= 0);
    userFile = fdopen(userFileFD, "rb");
    flock(userFileFD, LOCK_SH);
  }
}

void userFileClose() {
  flock(userFileFD, LOCK_UN);
  fclose(userFile);
}

int userFileRead(UserRec *ur) {
  return !fread(ur, sizeof(UserRec), 1, userFile);
}

int userFileWrite(UserRec *ur) {
  return !fwrite(ur, sizeof(UserRec), 1, userFile);
}

void userFileEdit(UserRec *ur) {
  fseek(userFile, -sizeof(UserRec), SEEK_CUR);
  userFileWrite(ur);
}

int userFileSearch(UserRec *ur, uid_t uid) {
  rewind(userFile);
  while (!userFileRead(ur))
    if (ur->uid == uid)
      return 0;
  return 1;
}

void banFileOpen() {
  if (geteuid())
    errQuit("must be root");
  banFileFD = open(banFilePath, O_CREAT | O_RDWR, 0644);
  assert(banFileFD >= 0);
  banFile = fdopen(banFileFD, "r+b");
  flock(banFileFD, LOCK_EX);
}

void banFileClose() {
  flock(banFileFD, LOCK_UN);
  fclose(banFile);
}

int banFileRead(BanRec *br) {
  return !fread(br, sizeof(BanRec), 1, banFile);
}

int banFileWrite(BanRec *br) {
  return !fwrite(br, sizeof(BanRec), 1, banFile);
}

int banFileSearch(BanRec *br, word phNoArea, word phNoLocal) {
  char phNoStr[11], phNoAreaStr[11], phNoLocalStr[11];
  sprintf(phNoAreaStr, "%u", phNoArea);
  sprintf(phNoLocalStr, "%u", phNoLocal);
  rewind(banFile);
  while (!banFileRead(br)) {
    if (br->flags) {
      sprintf(phNoStr, "%u", br->phNoArea);
      if ((phNoArea && !strncmp(phNoStr, phNoAreaStr, strlen(phNoStr))) ||
	  (phNoLocal && !strncmp(phNoStr, phNoLocalStr, strlen(phNoStr))) ||
	  (!phNoArea && !phNoLocal && !br->phNoArea && !br->phNoLocal))
	return 0;
    } else if (br->phNoArea == phNoArea && br->phNoLocal == phNoLocal)
      return 0;
  }
  return 1;
}

int phNoToWords(word *phNoArea, word *phNoLocal, char *phNo, int phNoDigits) {
  int i, len;
  char tPhNo[32], phNoAreaStr[32], phNoLocalStr[32];
  for (len = 0, i = 0; phNo[i] != '\0'; i++)
    if (isdigit(phNo[i]))
      tPhNo[len++] = phNo[i];
  tPhNo[len] = '\0';
  if (!len || len > 18) return 1;
  if (len < phNoDigits) {
    strcpy(phNoAreaStr, tPhNo);
    strcpy(phNoLocalStr, "0");
  } else {
    strcpy(phNoLocalStr, tPhNo + len - phNoDigits);
    strcpy(phNoAreaStr, tPhNo); phNoAreaStr[len - phNoDigits] = '\0';
  }
  *phNoArea = atol(phNoAreaStr);
  *phNoLocal = atol(phNoLocalStr);
  return 0;
}

char *wordsToPhNo(char *phNo, word phNoArea, word phNoLocal, char *phNoAreaFormat, char *phNoLocalFormat) {
  int i = 0, j = 0, k, l, skipChars;
  char phNoAreaStr[10], phNoLocalStr[10], *s;
  sprintf(phNoAreaStr, "%u", phNoArea);
  sprintf(phNoLocalStr, "%u", phNoLocal);
  if (!phNoLocal) {
    strcpy(phNo, phNoAreaStr);
    return phNo;
  }
  if (phNoArea) {
    s = phNoAreaFormat;
    skipChars = stringSum(phNoAreaFormat) - strlen(phNoAreaStr);
    while (*s != '\0') {
      if (isdigit(*s)) {
	l = strtol(s, &s, 10);
	if (skipChars) {
	  skipChars -= l;
	  if (skipChars == 0)
	    while (*s == ')' || *s == ' ') s++;
	  else if (skipChars < 0) goto err;
	  continue;
	}
	for (k = 0; k < l; k++)
	  phNo[i++] = phNoAreaStr[j++];
      } else while (!isdigit(*s) && *s != '\0')
	if (skipChars) s++;
	else phNo[i++] = *s++;
    }
    if (phNoAreaStr[j] != '\0') goto err;
    phNo[i++] = ' ';
  }
  if (phNoLocal) {
    j = 0;
    s = phNoLocalFormat;
    while (*s != '\0') {
      if (isdigit(*s)) {
	l = strtol(s, &s, 10);
	for (k = 0; k < l; k++)
	  phNo[i++] = phNoLocalStr[j++];
      } else while (!isdigit(*s) && *s != '\0')
	phNo[i++] = *s++;
    }
    if (phNoLocalStr[j] != '\0') goto err;
  }
  phNo[i] = '\0';
  return phNo;
err:
  phNo[0] = '\0';
  return NULL;
}

int stringSum(char *s) {
  int sum = 0;
  while (*s != '\0') {
    while (!isdigit(*s) && *s != '\0') s++;
    if (*s != '\0') sum += strtol(s, &s, 10);
  }
  return sum;
}

void bytes2ASCII(char *str, unsigned long long bytes) {
  if (bytes >= 1024 * 1024 * 1024) sprintf(str, "%.2f GB", (float)bytes / (1024 * 1024 * 1024));
  else if (bytes >= 1024 * 1024) sprintf(str, "%.2f MB", (float)bytes / (1024 * 1024));
  else sprintf(str, "%.2f KB", (float)bytes / 1024);
}
